@@ -18,44 +18,34 @@ import os
 import shutil
 import subprocess
 import sys
+import unittest
 
 from scripts.test.support import run_command, split_wast, node_test_glue, node_has_webassembly
 from scripts.test.shared import (
-    BIN_DIR, EMCC, MOZJS, NATIVECC, NATIVEXX, NODEJS, BINARYEN_JS,
-    WASM_AS, WASM_CTOR_EVAL, WASM_OPT, WASM_SHELL, WASM_MERGE, WASM_METADCE,
-    WASM_DIS, WASM_REDUCE, binary_format_check, delete_from_orbit, fail, fail_with_error,
+    BIN_DIR, MOZJS, NATIVECC, NATIVEXX, NODEJS, BINARYEN_JS, WASM_AS,
+    WASM_CTOR_EVAL, WASM_OPT, WASM_SHELL, WASM_METADCE, WASM_DIS, WASM_REDUCE,
+    binary_format_check, delete_from_orbit, fail, fail_with_error,
     fail_if_not_identical, fail_if_not_contained, has_vanilla_emcc,
-    has_vanilla_llvm, minify_check, num_failures, options, tests,
-    requested, warnings, has_shell_timeout, fail_if_not_identical_to_file
+    has_vanilla_llvm, minify_check, options, tests, requested, warnings,
+    has_shell_timeout, fail_if_not_identical_to_file, with_pass_debug
 )
 
-import scripts.test.asm2wasm as asm2wasm
-import scripts.test.lld as lld
-import scripts.test.wasm2js as wasm2js
+# For shared.num_failures. Cannot import directly because modifications made in
+# shared.py would not affect the version imported here.
+from scripts.test import shared
+from scripts.test import asm2wasm
+from scripts.test import lld
+from scripts.test import wasm2js
 
 if options.interpreter:
   print '[ using wasm interpreter at "%s" ]' % options.interpreter
   assert os.path.exists(options.interpreter), 'interpreter not found'
 
 
-# run a check with BINARYEN_PASS_DEBUG set, to do full validation
-def with_pass_debug(check):
-  old_pass_debug = os.environ.get('BINARYEN_PASS_DEBUG')
-  try:
-    os.environ['BINARYEN_PASS_DEBUG'] = '1'
-    check()
-  finally:
-    if old_pass_debug is not None:
-      os.environ['BINARYEN_PASS_DEBUG'] = old_pass_debug
-    else:
-      if 'BINARYEN_PASS_DEBUG' in os.environ:
-        del os.environ['BINARYEN_PASS_DEBUG']
-
-
 def run_help_tests():
   print '[ checking --help is useful... ]\n'
 
-  not_executable_suffix = ['.txt', '.js', '.ilk', '.pdb', '.dll']
+  not_executable_suffix = ['.txt', '.js', '.ilk', '.pdb', '.dll', '.wasm']
   executables = sorted(filter(lambda x: not any(x.endswith(s) for s in
                                                 not_executable_suffix) and os.path.isfile(x),
                               os.listdir(options.binaryen_bin)))
@@ -136,12 +126,12 @@ def run_wasm_opt_tests():
     if t.endswith('.wast'):
       print '..', t
       wasm = os.path.basename(t).replace('.wast', '')
-      cmd = WASM_OPT + [os.path.join(options.binaryen_test, 'print', t), '--print']
+      cmd = WASM_OPT + [os.path.join(options.binaryen_test, 'print', t), '--print', '-all']
       print '    ', ' '.join(cmd)
       actual, err = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, universal_newlines=True).communicate()
       expected_file = os.path.join(options.binaryen_test, 'print', wasm + '.txt')
       fail_if_not_identical_to_file(actual, expected_file)
-      cmd = WASM_OPT + [os.path.join(options.binaryen_test, 'print', t), '--print-minified']
+      cmd = WASM_OPT + [os.path.join(options.binaryen_test, 'print', t), '--print-minified', '-all']
       print '    ', ' '.join(cmd)
       actual, err = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, universal_newlines=True).communicate()
       fail_if_not_identical(actual.strip(), open(os.path.join(options.binaryen_test, 'print', wasm + '.minified.txt')).read().strip())
@@ -153,7 +143,7 @@ def run_wasm_opt_tests():
       print '..', t
       t = os.path.join(options.binaryen_test, t)
       f = t + '.from-wast'
-      cmd = WASM_OPT + [t, '--print']
+      cmd = WASM_OPT + [t, '--print', '-all']
       actual = run_command(cmd)
       actual = actual.replace('printing before:\n', '')
 
@@ -193,39 +183,12 @@ def run_wasm_dis_tests():
 
       # also verify there are no validation errors
       def check():
-        cmd = WASM_OPT + [t]
+        cmd = WASM_OPT + [t, '-all']
         run_command(cmd)
 
       with_pass_debug(check)
 
 
-def run_wasm_merge_tests():
-  print '\n[ checking wasm-merge... ]\n'
-
-  test_dir = os.path.join(options.binaryen_test, 'merge')
-  for t in os.listdir(test_dir):
-    if t.endswith(('.wast', '.wasm')):
-      print '..', t
-      t = os.path.join(test_dir, t)
-      u = t + '.toMerge'
-      for finalize in [0, 1]:
-        for opt in [0, 1]:
-          cmd = WASM_MERGE + [t, u, '-o', 'a.wast', '-S', '--verbose']
-          if finalize:
-            cmd += ['--finalize-memory-base=1024', '--finalize-table-base=8']
-          if opt:
-            cmd += ['-O']
-          stdout = run_command(cmd)
-          actual = open('a.wast').read()
-          out = t + '.combined'
-          if finalize:
-            out += '.finalized'
-          if opt:
-            out += '.opt'
-          fail_if_not_identical_to_file(actual, out)
-          fail_if_not_identical_to_file(stdout, out + '.stdout')
-
-
 def run_crash_tests():
   print "\n[ checking we don't crash on tricky inputs... ]\n"
 
@@ -298,7 +261,7 @@ def run_wasm_reduce_tests():
       t = os.path.join(test_dir, t)
       # convert to wasm
       run_command(WASM_AS + [t, '-o', 'a.wasm'])
-      run_command(WASM_REDUCE + ['a.wasm', '--command=%s b.wasm --fuzz-exec' % WASM_OPT[0], '-t', 'b.wasm', '-w', 'c.wasm', '--timeout=4'])
+      run_command(WASM_REDUCE + ['a.wasm', '--command=%s b.wasm --fuzz-exec -all' % WASM_OPT[0], '-t', 'b.wasm', '-w', 'c.wasm', '--timeout=4'])
       expected = t + '.txt'
       run_command(WASM_DIS + ['c.wasm', '-o', 'a.wast'])
       with open('a.wast') as seen:
@@ -309,11 +272,11 @@ def run_wasm_reduce_tests():
   if 'fsanitize=thread' not in str(os.environ):
     print '\n[ checking wasm-reduce fuzz testcase ]\n'
 
-    run_command(WASM_OPT + [os.path.join(options.binaryen_test, 'unreachable-import_wasm-only.asm.js'), '-ttf', '-Os', '-o', 'a.wasm'])
+    run_command(WASM_OPT + [os.path.join(options.binaryen_test, 'unreachable-import_wasm-only.asm.js'), '-ttf', '-Os', '-o', 'a.wasm', '-all'])
     before = os.stat('a.wasm').st_size
-    run_command(WASM_REDUCE + ['a.wasm', '--command=%s b.wasm --fuzz-exec' % WASM_OPT[0], '-t', 'b.wasm', '-w', 'c.wasm'])
+    run_command(WASM_REDUCE + ['a.wasm', '--command=%s b.wasm --fuzz-exec -all' % WASM_OPT[0], '-t', 'b.wasm', '-w', 'c.wasm'])
     after = os.stat('c.wasm').st_size
-    assert after < 0.333 * before, [before, after]
+    assert after < 0.6 * before, [before, after]
 
 
 def run_spec_tests():
@@ -321,7 +284,7 @@ def run_spec_tests():
 
   if len(requested) == 0:
     # FIXME we support old and new memory formats, for now, until 0xc, and so can't pass this old-style test.
-    BLACKLIST = ['memory.wast', 'binary.wast']
+    BLACKLIST = ['binary.wast']
     # FIXME to update the spec to 0xd, we need to implement (register "name") for import.wast
     spec_tests = [os.path.join('spec', t) for t in sorted(os.listdir(os.path.join(options.binaryen_test, 'spec'))) if t not in BLACKLIST]
   else:
@@ -346,7 +309,7 @@ def run_spec_tests():
 
       def run_opt_test(wast):
         # check optimization validation
-        cmd = WASM_OPT + [wast, '-O']
+        cmd = WASM_OPT + [wast, '-O', '-all']
         run_command(cmd)
 
       def check_expected(actual, expected):
@@ -354,7 +317,7 @@ def run_spec_tests():
           expected = open(expected).read()
 
           # fix it up, our pretty (i32.const 83) must become compared to a homely 83 : i32
-          def fix(x):
+          def fix_expected(x):
             x = x.strip()
             if not x:
               return x
@@ -363,7 +326,13 @@ def run_spec_tests():
               v = v[:-1]  # remove trailing '.'
             return '(' + t + '.const ' + v + ')'
 
-          expected = '\n'.join(map(fix, expected.split('\n')))
+          def fix_actual(x):
+            if '[trap ' in x:
+              return ''
+            return x
+
+          expected = '\n'.join(map(fix_expected, expected.split('\n')))
+          actual = '\n'.join(map(fix_actual, actual.split('\n')))
           print '       (using expected output)'
           actual = actual.strip()
           expected = expected.strip()
@@ -586,64 +555,15 @@ def run_gcc_tests():
       fail_if_not_identical_to_file(actual, expected)
 
 
-def run_emscripten_tests():
-  if not os.path.exists(os.path.join(options.binaryen_bin, 'wasm.js')):
-    print 'no wasm.js build to test'
-    return
+def run_unittest():
+  print '\n[ checking unit tests...]\n'
 
-  print '\n[ checking wasm.js methods... ]\n'
-
-  for method_init in ['interpret-asm2wasm', 'interpret-s-expr', 'asmjs', 'interpret-binary', 'asmjs,interpret-binary', 'interpret-binary,asmjs']:
-    # check success and failure for simple modes, only success for combined/fallback ones
-    for success in [1, 0] if ',' not in method_init else [1]:
-      method = method_init
-      command = [EMCC, '-o', 'a.wasm.js', '-s', 'BINARYEN=1', os.path.join(options.binaryen_test, 'hello_world.c')]
-      command += ['-s', 'BINARYEN_METHOD="' + method + '"']
-      print method, ' : ', ' '.join(command), ' => ', success
-      subprocess.check_call(command)
-
-      see_polyfill = 'var WasmJS = ' in open('a.wasm.js').read()
-
-      if method and 'interpret' not in method:
-        assert not see_polyfill, 'verify polyfill was not added - we specified a method, and it does not need it'
-      else:
-        assert see_polyfill, 'we need the polyfill'
-
-      def break_cashew():
-        asm = open('a.wasm.asm.js').read()
-        asm = asm.replace('"almost asm"', '"use asm"; var not_in_asm = [].length + (true || { x: 5 }.x);')
-        asm = asm.replace("'almost asm'", '"use asm"; var not_in_asm = [].length + (true || { x: 5 }.x);')
-        with open('a.wasm.asm.js', 'w') as o:
-          o.write(asm)
-      if method.startswith('interpret-asm2wasm'):
-        delete_from_orbit('a.wasm.wast')  # we should not need the .wast
-        if not success:
-          break_cashew()  # we need cashew
-      elif method.startswith('interpret-s-expr'):
-        delete_from_orbit('a.wasm.asm.js')  # we should not need the .asm.js
-        if not success:
-          delete_from_orbit('a.wasm.wast')
-      elif method.startswith('asmjs'):
-        delete_from_orbit('a.wasm.wast')  # we should not need the .wast
-        break_cashew()  # we don't use cashew, so ok to break it
-        if not success:
-          delete_from_orbit('a.wasm.js')
-      elif method.startswith('interpret-binary'):
-        delete_from_orbit('a.wasm.wast')  # we should not need the .wast
-        delete_from_orbit('a.wasm.asm.js')  # we should not need the .asm.js
-        if not success:
-          delete_from_orbit('a.wasm.wasm')
-      else:
-        1 / 0
-      if NODEJS:
-        proc = subprocess.Popen([NODEJS, 'a.wasm.js'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
-        out, err = proc.communicate()
-        if success:
-          assert proc.returncode == 0, err
-          assert 'hello, world!' in out, out
-        else:
-          assert proc.returncode != 0, err
-          assert 'hello, world!' not in out, out
+  # equivalent to `python -m unittest discover -s ./test -v`
+  suite = unittest.defaultTestLoader.discover(os.path.dirname(options.binaryen_test))
+  result = unittest.TextTestRunner(verbosity=2, failfast=options.abort_on_first_failure).run(suite)
+  shared.num_failures += len(result.errors) + len(result.failures)
+  if options.abort_on_first_failure and shared.num_failures:
+    raise Exception("unittest failed")
 
 
 # Run all the tests
@@ -653,7 +573,6 @@ def main():
   asm2wasm.test_asm2wasm()
   asm2wasm.test_asm2wasm_binary()
   run_wasm_dis_tests()
-  run_wasm_merge_tests()
   run_crash_tests()
   run_dylink_tests()
   run_ctor_eval_tests()
@@ -671,20 +590,21 @@ def main():
   print '\n[ checking example testcases... ]\n'
   if options.run_gcc_tests:
     run_gcc_tests()
-  if EMCC:
-    run_emscripten_tests()
+
+  run_unittest()
 
   # Check/display the results
-  if num_failures == 0:
+  if shared.num_failures == 0:
     print '\n[ success! ]'
 
   if warnings:
     print '\n' + '\n'.join(warnings)
 
-  if num_failures > 0:
-    print '\n[ ' + str(num_failures) + ' failures! ]'
+  if shared.num_failures > 0:
+    print '\n[ ' + str(shared.num_failures) + ' failures! ]'
+    return 1
 
-  return num_failures
+  return 0
 
 
 if __name__ == '__main__':
