@@ -80,7 +80,7 @@ class GlobalMixedArena : public MixedArena {
 public:
   template<class T>
   T* alloc() {
-    auto* ret = static_cast<T*>(allocSpace(sizeof(T)));
+    auto* ret = static_cast<T*>(allocSpace(sizeof(T), alignof(T)));
     new (ret) T();
     return ret;
   }
@@ -92,7 +92,7 @@ class ArrayStorage : public ArenaVectorBase<ArrayStorage, Ref> {
 public:
   void allocate(size_t size) {
     allocatedElements = size;
-    data = static_cast<Ref*>(arena.allocSpace(sizeof(Ref) * allocatedElements));
+    data = static_cast<Ref*>(arena.allocSpace(sizeof(Ref) * allocatedElements, alignof(Ref)));
   }
 };
 
@@ -112,7 +112,7 @@ struct Value {
     AssignName_ = 7
   };
 
-  Type type;
+  Type type = Null;
 
   typedef std::unordered_map<IString, Ref> ObjectStorage;
 
@@ -131,14 +131,14 @@ struct Value {
   };
 
   // constructors all copy their input
-  Value() : type(Null), num(0) {}
-  explicit Value(const char *s) : type(Null) {
+  Value() {}
+  explicit Value(const char *s) {
     setString(s);
   }
-  explicit Value(double n) : type(Null) {
+  explicit Value(double n) {
     setNumber(n);
   }
-  explicit Value(ArrayStorage &a) : type(Null) {
+  explicit Value(ArrayStorage &a) {
     setArray();
     *arr = a;
   }
@@ -544,15 +544,16 @@ void traverseFunctions(Ref ast, std::function<void (Ref)> visit);
 struct JSPrinter {
   bool pretty, finalize;
 
-  char *buffer;
-  size_t size, used;
+  char *buffer = nullptr;
+  size_t size = 0;
+  size_t used = 0;
 
-  int indent;
-  bool possibleSpace; // add a space to separate identifiers
+  int indent = 0;
+  bool possibleSpace = false; // add a space to separate identifiers
 
   Ref ast;
 
-  JSPrinter(bool pretty_, bool finalize_, Ref ast_) : pretty(pretty_), finalize(finalize_), buffer(0), size(0), used(0), indent(0), possibleSpace(false), ast(ast_) {}
+  JSPrinter(bool pretty_, bool finalize_, Ref ast_) : pretty(pretty_), finalize(finalize_), ast(ast_) {}
 
   ~JSPrinter() {
     free(buffer);
@@ -560,6 +561,7 @@ struct JSPrinter {
 
   void printAst() {
     print(ast);
+    ensure(1);
     buffer[used] = 0;
   }
 
@@ -632,8 +634,13 @@ struct JSPrinter {
     return node->isArray() && node[0] == DEFUN;
   }
 
-  bool isBlock(Ref node) {
-    return node->isArray() && node[0] == BLOCK;
+  bool endsInBlock(Ref node) {
+    if (node->isArray() && node[0] == BLOCK) return true;
+    // Check for a label on a block
+    if (node->isArray() && node[0] == LABEL && endsInBlock(node[2])) return true;
+    // Check for an if
+    if (node->isArray() && node[0] == IF && endsInBlock(ifHasElse(node) ? node[3] : node[2])) return true;
+    return false;
   }
 
   bool isIf(Ref node) {
@@ -761,7 +768,7 @@ struct JSPrinter {
         if (first) first = false;
         else newline();
         print(curr);
-        if (!isDefun(curr) && !isBlock(curr) && !isIf(curr)) {
+        if (!isDefun(curr) && !endsInBlock(curr) && !isIf(curr)) {
           emit(';');
         }
       }
@@ -979,6 +986,9 @@ struct JSPrinter {
   }
 
   void printNum(Ref node) {
+    if (node->getNumber() < 0 && buffer[used-1] == '-') {
+      emit(' '); // cannot join - and - to --, looks like the -- operator
+    }
     emit(numToString(node->getNumber(), finalize));
   }
 
@@ -1208,42 +1218,24 @@ struct JSPrinter {
     print(node[1]);
     emit(')');
     space();
-    // special case: we need braces to save us from ambiguity, if () { if () } else. otherwise else binds to inner if
-    // also need to recurse for                                if () { if () { } else { if () } else
-    // (note that this is only a problem if the if body has a single element in it, not a block or such, as then
-    // the block would be braced)
-    // this analysis is a little conservative - it assumes any child if could be confused with us, which implies
-    // all other braces vanished (the worst case for us, we are not saved by other braces).
-    bool needBraces = false;
-    bool hasElse = ifHasElse(node);
-    if (hasElse) {
-      Ref child = node[2];
-      while (child->isArray() && child[0] == IF) {
-        if (!ifHasElse(child)) {
-          needBraces = true;
-          break;
-        }
-        child = child[3]; // continue into the else
-      }
-    }
-    if (needBraces) {
-      emit('{');
-      indent++;
-      newline();
-      print(node[2]);
-      indent--;
-      newline();
-      emit('}');
-    } else {
-      print(node[2], "{}");
-      if (!isBlock(node[2])) emit(';');
-    }
-    if (hasElse) {
+    emit('{');
+    indent++;
+    newline();
+    print(node[2]);
+    indent--;
+    newline();
+    emit('}');
+    if (ifHasElse(node)) {
       space();
       emit("else");
       safeSpace();
-      print(node[3], "{}");
-      if (!isBlock(node[3])) emit(';');
+      emit('{');
+      indent++;
+      newline();
+      print(node[3]);
+      indent--;
+      newline();
+      emit('}');
     }
   }
 
@@ -1326,9 +1318,18 @@ struct JSPrinter {
         pretty ? emit(", ") : emit(',');
         newline();
       }
-      const char *str = args[i][0]->getCString();
-      const char *check = str;
       bool needQuote = false;
+      const char *str;
+      if (args[i][0]->isArray()) {
+        assert(args[i][0][0] == STRING);
+        // A quoted string.
+        needQuote = true;
+        str = args[i][0][1]->getCString();
+      } else {
+        // Just a raw string, no quotes.
+        str = args[i][0]->getCString();
+      }
+      const char *check = str;
       while (*check) {
         if (!isalnum(*check) && *check != '_' && *check != '$') {
           needQuote = true;
@@ -1652,6 +1653,12 @@ public:
                                          .push_back(value));
   }
 
+  static void appendToObjectWithQuotes(Ref array, IString key, Ref value) {
+    assert(array[0] == OBJECT);
+    array[1]->push_back(&makeRawArray(2)->push_back(makeString(key))
+                                         .push_back(value));
+  }
+
   static Ref makeSub(Ref obj, Ref index) {
     return &makeRawArray(2)->push_back(makeRawString(SUB))
                             .push_back(obj)
