@@ -27,6 +27,10 @@
 #include "emscripten-optimizer/istring.h"
 #include "support/name.h"
 #include "wasm-builder.h"
+#include "abi/js.h"
+#include "ir/flat.h"
+#include "ir/iteration.h"
+#include "ir/memory-utils.h"
 #include "ir/module-utils.h"
 #include "ir/names.h"
 #include "asmjs/shared-constants.h"
@@ -34,9 +38,7 @@
 namespace wasm {
 
 static Name makeHighName(Name n) {
-  return Name(
-    cashew::IString((std::string(n.c_str()) + "$hi").c_str(), false)
-  );
+  return std::string(n.c_str()) + "$hi";
 }
 
 struct I64ToI32Lowering : public WalkerPass<PostWalker<I64ToI32Lowering>> {
@@ -102,12 +104,27 @@ struct I64ToI32Lowering : public WalkerPass<PostWalker<I64ToI32Lowering>> {
     if (!builder) builder = make_unique<Builder>(*module);
     // add new globals for high bits
     for (size_t i = 0, globals = module->globals.size(); i < globals; ++i) {
-      auto& curr = module->globals[i];
+      auto* curr = module->globals[i].get();
       if (curr->type != i64) continue;
+      originallyI64Globals.insert(curr->name);
       curr->type = i32;
-      auto* high = ModuleUtils::copyGlobal(curr.get(), *module);
-      high->name = makeHighName(curr->name);
+      auto* high = builder->makeGlobal(makeHighName(curr->name), i32, builder->makeConst(Literal(int32_t(0))), Builder::Mutable);
       module->addGlobal(high);
+      if (curr->imported()) {
+        Fatal() << "TODO: imported i64 globals";
+      } else {
+        if (auto* c = curr->init->dynCast<Const>()) {
+          uint64_t value = c->value.geti64();
+          c->value = Literal(uint32_t(value));
+          c->type = i32;
+          high->init = builder->makeConst(Literal(uint32_t(value >> 32)));
+        } else if (auto* get = curr->init->dynCast<GetGlobal>()) {
+          high->init = builder->makeGetGlobal(makeHighName(get->name), i32);
+        } else {
+          WASM_UNREACHABLE();
+        }
+        curr->init->type = i32;
+      }
     }
 
     // For functions that return 64-bit values, we use this global variable
@@ -138,11 +155,11 @@ struct I64ToI32Lowering : public WalkerPass<PostWalker<I64ToI32Lowering>> {
   }
 
   void doWalkFunction(Function* func) {
+    Flat::verifyFlatness(func);
     // create builder here if this is first entry to module for this object
     if (!builder) builder = make_unique<Builder>(*getModule());
     indexMap.clear();
     highBitVars.clear();
-    labelHighBitVars.clear();
     freeTemps.clear();
     Module temp;
     auto* oldFunc = ModuleUtils::copyFunction(func, temp);
@@ -203,148 +220,33 @@ struct I64ToI32Lowering : public WalkerPass<PostWalker<I64ToI32Lowering>> {
     }
   }
 
-  void visitBlock(Block* curr) {
-    if (curr->list.size() == 0) return;
-    if (curr->type == i64) curr->type = i32;
-    auto highBitsIt = labelHighBitVars.find(curr->name);
-    if (!hasOutParam(curr->list.back())) {
-      if (highBitsIt != labelHighBitVars.end()) {
-        setOutParam(curr, std::move(highBitsIt->second));
-      }
-      return;
-    }
-    TempVar lastHighBits = fetchOutParam(curr->list.back());
-    if (highBitsIt == labelHighBitVars.end() ||
-        highBitsIt->second == lastHighBits) {
-      setOutParam(curr, std::move(lastHighBits));
-      if (highBitsIt != labelHighBitVars.end()) {
-        labelHighBitVars.erase(highBitsIt);
-      }
-      return;
-    }
-    TempVar highBits = std::move(highBitsIt->second);
-    TempVar tmp = getTemp();
-    labelHighBitVars.erase(highBitsIt);
-    SetLocal* setLow = builder->makeSetLocal(tmp, curr->list.back());
-    SetLocal* setHigh = builder->makeSetLocal(
-      highBits,
-      builder->makeGetLocal(lastHighBits, i32)
-    );
-    GetLocal* getLow = builder->makeGetLocal(tmp, i32);
-    curr->list.back() = builder->blockify(setLow, setHigh, getLow);
-    setOutParam(curr, std::move(highBits));
-  }
-
-  // If and Select have identical code
-  template <typename T>
-  void visitBranching(T* curr) {
-    if (!hasOutParam(curr->ifTrue)) return;
-    assert(curr->ifFalse != nullptr && "Nullable ifFalse found");
-    TempVar highBits = fetchOutParam(curr->ifTrue);
-    TempVar falseBits = fetchOutParam(curr->ifFalse);
-    TempVar tmp = getTemp();
-    curr->type = i32;
-    curr->ifFalse = builder->blockify(
-      builder->makeSetLocal(tmp, curr->ifFalse),
-      builder->makeSetLocal(
-        highBits,
-        builder->makeGetLocal(falseBits, i32)
-      ),
-      builder->makeGetLocal(tmp, i32)
-    );
-    setOutParam(curr, std::move(highBits));
-  }
-
-  void visitIf(If* curr) {
-    visitBranching<If>(curr);
-  }
-
-  void visitLoop(Loop* curr) {
-    assert(labelHighBitVars.find(curr->name) == labelHighBitVars.end());
-    if (curr->type != i64) return;
-    curr->type = i32;
-    setOutParam(curr, fetchOutParam(curr->body));
-  }
-
-  void visitBreak(Break* curr) {
-    if (!hasOutParam(curr->value)) return;
-    assert(curr->value != nullptr);
-    TempVar valHighBits = fetchOutParam(curr->value);
-    auto blockHighBitsIt = labelHighBitVars.find(curr->name);
-    if (blockHighBitsIt == labelHighBitVars.end()) {
-      labelHighBitVars.emplace(curr->name, std::move(valHighBits));
-      curr->type = i32;
-      return;
-    }
-    TempVar blockHighBits = std::move(blockHighBitsIt->second);
-    TempVar tmp = getTemp();
-    SetLocal* setLow = builder->makeSetLocal(tmp, curr->value);
-    SetLocal* setHigh = builder->makeSetLocal(
-      blockHighBits,
-      builder->makeGetLocal(valHighBits, i32)
-    );
-    curr->value = builder->makeGetLocal(tmp, i32);
-    curr->type = i32;
-    replaceCurrent(builder->blockify(setLow, setHigh, curr));
-  }
-
-  void visitSwitch(Switch* curr) {
-    if (!hasOutParam(curr->value)) return;
-    TempVar outParam = fetchOutParam(curr->value);
-    TempVar tmp = getTemp();
-    Expression* result = curr;
-    std::vector<Name> targets;
-    size_t blockID = 0;
-    auto processTarget = [&](Name target) -> Name {
-      auto labelIt = labelHighBitVars.find(target);
-      if (labelIt == labelHighBitVars.end()) {
-        labelHighBitVars.emplace(target, getTemp());
-        labelIt = labelHighBitVars.find(target);
-      }
-      Name newLabel("$i64toi32_" + std::string(target.c_str()) +
-                    "_" + std::to_string(blockID++));
-      Block* trampoline = builder->makeBlock(newLabel, result);
-      trampoline->type = i32;
-      result = builder->blockify(
-        builder->makeSetLocal(tmp, trampoline),
-        builder->makeSetLocal(
-          labelIt->second,
-          builder->makeGetLocal(outParam, i32)
-        ),
-        builder->makeBreak(target, builder->makeGetLocal(tmp, i32))
-      );
-      return newLabel;
-    };
-    for (auto target : curr->targets) {
-      targets.push_back(processTarget(target));
-    }
-    curr->targets.set(targets);
-    curr->default_ = processTarget(curr->default_);
-    replaceCurrent(result);
-  }
-
-  template <typename T>
+  template<typename T>
   using BuilderFunc = std::function<T*(std::vector<Expression*>&, Type)>;
 
-  template <typename T>
-  void visitGenericCall(T* curr, BuilderFunc<T> callBuilder) {
+  // Fixes up a call. If we performed fixups, returns the call; otherwise returns nullptr;
+  template<typename T>
+  T* visitGenericCall(T* curr, BuilderFunc<T> callBuilder) {
+    bool fixed = false;
     std::vector<Expression*> args;
     for (auto* e : curr->operands) {
       args.push_back(e);
       if (hasOutParam(e)) {
         TempVar argHighBits = fetchOutParam(e);
         args.push_back(builder->makeGetLocal(argHighBits, i32));
+        fixed =  true;
       }
     }
     if (curr->type != i64) {
-      replaceCurrent(callBuilder(args, curr->type));
-      return;
+      auto* ret = callBuilder(args, curr->type);
+      replaceCurrent(ret);
+      return fixed ? ret : nullptr;
     }
     TempVar lowBits = getTemp();
     TempVar highBits = getTemp();
+    auto* call = callBuilder(args, i32);
     SetLocal* doCall = builder->makeSetLocal(
       lowBits,
-      callBuilder(args, i32)
+      call
     );
     SetLocal* setHigh = builder->makeSetLocal(
       highBits,
@@ -354,14 +256,21 @@ struct I64ToI32Lowering : public WalkerPass<PostWalker<I64ToI32Lowering>> {
     Block* result = builder->blockify(doCall, setHigh, getLow);
     setOutParam(result, std::move(highBits));
     replaceCurrent(result);
+    return call;
   }
   void visitCall(Call* curr) {
-    visitGenericCall<Call>(
+    auto* fixedCall = visitGenericCall<Call>(
       curr,
       [&](std::vector<Expression*>& args, Type ty) {
         return builder->makeCall(curr->target, args, ty);
       }
     );
+    // If this was to an import, we need to call the legal version. This assumes
+    // that legalize-js-interface has been run before.
+    if (fixedCall && getModule()->getFunction(fixedCall->target)->imported()) {
+      fixedCall->target = std::string("legalfunc$") + fixedCall->target.str;
+      return;
+    }
   }
 
   void visitCallIndirect(CallIndirect* curr) {
@@ -428,7 +337,7 @@ struct I64ToI32Lowering : public WalkerPass<PostWalker<I64ToI32Lowering>> {
       return;
     }
     TempVar highBits = fetchOutParam(curr->value);
-    SetLocal* setHigh = builder->makeSetLocal(
+    auto* setHigh = builder->makeSetLocal(
       mappedIndex + 1,
       builder->makeGetLocal(highBits, i32)
     );
@@ -437,13 +346,31 @@ struct I64ToI32Lowering : public WalkerPass<PostWalker<I64ToI32Lowering>> {
   }
 
   void visitGetGlobal(GetGlobal* curr) {
-    if (curr->type != i64) return;
-    assert(false && "GetGlobal not implemented");
+    if (!getFunction()) return; // if in a global init, skip - we already handled that.
+    if (!originallyI64Globals.count(curr->name)) return;
+    curr->type = i32;
+    TempVar highBits = getTemp();
+    SetLocal *setHighBits = builder->makeSetLocal(
+      highBits,
+      builder->makeGetGlobal(
+        makeHighName(curr->name),
+        i32
+      )
+    );
+    Block* result = builder->blockify(setHighBits, curr);
+    replaceCurrent(result);
+    setOutParam(result, std::move(highBits));
   }
 
   void visitSetGlobal(SetGlobal* curr) {
-    if (curr->type != i64) return;
-    assert(false && "SetGlobal not implemented");
+    if (!originallyI64Globals.count(curr->name)) return;
+    if (handleUnreachable(curr)) return;
+    TempVar highBits = fetchOutParam(curr->value);
+    auto* setHigh = builder->makeSetGlobal(
+      makeHighName(curr->name),
+      builder->makeGetLocal(highBits, i32)
+    );
+    replaceCurrent(builder->makeSequence(curr, setHigh));
   }
 
   void visitLoad(Load* curr) {
@@ -508,6 +435,7 @@ struct I64ToI32Lowering : public WalkerPass<PostWalker<I64ToI32Lowering>> {
       TempVar ptrTemp = getTemp();
       SetLocal* setPtr = builder->makeSetLocal(ptrTemp, curr->ptr);
       curr->ptr = builder->makeGetLocal(ptrTemp, i32);
+      curr->finalize();
       Store* storeHigh = builder->makeStore(
         4,
         curr->offset + 4,
@@ -529,6 +457,7 @@ struct I64ToI32Lowering : public WalkerPass<PostWalker<I64ToI32Lowering>> {
   }
 
   void visitConst(Const* curr) {
+    if (!getFunction()) return; // if in a global init, skip - we already handled that.
     if (curr->type != i64) return;
     TempVar highBits = getTemp();
     Const* lowVal = builder->makeConst(
@@ -605,15 +534,17 @@ struct I64ToI32Lowering : public WalkerPass<PostWalker<I64ToI32Lowering>> {
     // our f64 through memory at address 0
     TempVar highBits = getTemp();
     Block *result = builder->blockify(
-      builder->makeStore(8, 0, 8, builder->makeConst(Literal(int32_t(0))), curr->value, f64),
+      builder->makeCall(ABI::wasm2js::SCRATCH_STORE_F64, { curr->value }, none),
       builder->makeSetLocal(
         highBits,
-        builder->makeLoad(4, true, 4, 4, builder->makeConst(Literal(int32_t(0))), i32)
+        builder->makeCall(ABI::wasm2js::SCRATCH_LOAD_I32, { builder->makeConst(Literal(int32_t(1))) }, i32)
       ),
-      builder->makeLoad(4, true, 0, 4, builder->makeConst(Literal(int32_t(0))), i32)
+      builder->makeCall(ABI::wasm2js::SCRATCH_LOAD_I32, { builder->makeConst(Literal(int32_t(0))) }, i32)
     );
     setOutParam(result, std::move(highBits));
     replaceCurrent(result);
+    MemoryUtils::ensureExists(getModule()->memory);
+    ABI::wasm2js::ensureScratchMemoryHelpers(getModule());
   }
 
   void lowerReinterpretInt64(Unary* curr) {
@@ -621,11 +552,13 @@ struct I64ToI32Lowering : public WalkerPass<PostWalker<I64ToI32Lowering>> {
     // our i64 through memory at address 0
     TempVar highBits = fetchOutParam(curr->value);
     Block *result = builder->blockify(
-      builder->makeStore(4, 0, 4, builder->makeConst(Literal(int32_t(0))), curr->value, i32),
-      builder->makeStore(4, 4, 4, builder->makeConst(Literal(int32_t(0))), builder->makeGetLocal(highBits, i32), i32),
-      builder->makeLoad(8, true, 0, 8, builder->makeConst(Literal(int32_t(0))), f64)
+      builder->makeCall(ABI::wasm2js::SCRATCH_STORE_I32, { builder->makeConst(Literal(int32_t(0))), curr->value }, none),
+      builder->makeCall(ABI::wasm2js::SCRATCH_STORE_I32, { builder->makeConst(Literal(int32_t(1))), builder->makeGetLocal(highBits, i32) }, none),
+      builder->makeCall(ABI::wasm2js::SCRATCH_LOAD_F64, {}, f64)
     );
     replaceCurrent(result);
+    MemoryUtils::ensureExists(getModule()->memory);
+    ABI::wasm2js::ensureScratchMemoryHelpers(getModule());
   }
 
   void lowerTruncFloatToInt(Unary *curr) {
@@ -894,11 +827,7 @@ struct I64ToI32Lowering : public WalkerPass<PostWalker<I64ToI32Lowering>> {
 
   void visitUnary(Unary* curr) {
     if (!unaryNeedsLowering(curr->op)) return;
-    if (curr->type == unreachable || curr->value->type == unreachable) {
-      assert(!hasOutParam(curr->value));
-      replaceCurrent(curr->value);
-      return;
-    }
+    if (handleUnreachable(curr)) return;
     assert(hasOutParam(curr->value) || curr->type == i64 || curr->type == f64);
     switch (curr->op) {
       case ClzInt64:
@@ -1416,25 +1345,8 @@ struct I64ToI32Lowering : public WalkerPass<PostWalker<I64ToI32Lowering>> {
   }
 
   void visitBinary(Binary* curr) {
+    if (handleUnreachable(curr)) return;
     if (!binaryNeedsLowering(curr->op)) return;
-    if (!hasOutParam(curr->left)) {
-      // left unreachable, replace self with left
-      replaceCurrent(curr->left);
-      if (hasOutParam(curr->right)) {
-        // free temp var
-        fetchOutParam(curr->right);
-      }
-      return;
-    }
-    if (!hasOutParam(curr->right)) {
-      // right unreachable, replace self with left then right
-      replaceCurrent(
-        builder->blockify(builder->makeDrop(curr->left), curr->right)
-      );
-      // free temp var
-      fetchOutParam(curr->left);
-      return;
-    }
     // left and right reachable, lower normally
     TempVar leftLow = getTemp();
     TempVar leftHigh = fetchOutParam(curr->left);
@@ -1526,7 +1438,37 @@ struct I64ToI32Lowering : public WalkerPass<PostWalker<I64ToI32Lowering>> {
   }
 
   void visitSelect(Select* curr) {
-    visitBranching<Select>(curr);
+    if (handleUnreachable(curr)) return;
+    if (!hasOutParam(curr->ifTrue)) {
+      assert(!hasOutParam(curr->ifFalse));
+      return;
+    }
+    assert(hasOutParam(curr->ifFalse));
+    TempVar highBits = getTemp();
+    TempVar lowBits = getTemp();
+    TempVar cond = getTemp();
+    Block* result = builder->blockify(
+      builder->makeSetLocal(cond, curr->condition),
+      builder->makeSetLocal(
+        lowBits,
+        builder->makeSelect(
+          builder->makeGetLocal(cond, i32),
+          curr->ifTrue,
+          curr->ifFalse
+        )
+      ),
+      builder->makeSetLocal(
+        highBits,
+        builder->makeSelect(
+          builder->makeGetLocal(cond, i32),
+          builder->makeGetLocal(fetchOutParam(curr->ifTrue), i32),
+          builder->makeGetLocal(fetchOutParam(curr->ifFalse), i32)
+        )
+      ),
+      builder->makeGetLocal(lowBits, i32)
+    );
+    setOutParam(result, std::move(highBits));
+    replaceCurrent(result);
   }
 
   void visitDrop(Drop* curr) {
@@ -1554,8 +1496,8 @@ private:
   std::unordered_map<Index, Index> indexMap;
   std::unordered_map<int, std::vector<Index>> freeTemps;
   std::unordered_map<Expression*, TempVar> highBitVars;
-  std::unordered_map<Name, TempVar> labelHighBitVars;
   std::unordered_map<Index, Type> tempTypes;
+  std::unordered_set<Name> originallyI64Globals;
   Index nextTemp;
 
   TempVar getTemp(Type ty = i32) {
@@ -1587,6 +1529,34 @@ private:
     highBitVars.erase(e);
     return ret;
   }
+
+  // If e.g. a select is unreachable, then one arm may have an out param
+  // but not the other. In this case dce should really have been run
+  // before; handle it in a simple way here by replacing the node with
+  // a block of its children.
+  // This is valid only for nodes that execute their children
+  // unconditionally before themselves, so it is not valid for an if,
+  // in particular.
+  bool handleUnreachable(Expression* curr) {
+    if (curr->type != unreachable) return false;
+    std::vector<Expression*> children;
+    bool hasUnreachable = false;
+    for (auto* child : ChildIterator(curr)) {
+      if (isConcreteType(child->type)) {
+        child = builder->makeDrop(child);
+      } else if (child->type == unreachable) {
+        hasUnreachable = true;
+      }
+      children.push_back(child);
+    }
+    if (!hasUnreachable) return false;
+    // This has an unreachable child, so we can replace it with
+    // the children.
+    auto* block = builder->makeBlock(children);
+    assert(block->type == unreachable);
+    replaceCurrent(block);
+    return true;
+  }
 };
 
 Pass *createI64ToI32LoweringPass() {
